use std::ffi::OsString;
use std::fs::File;
use std::io::Write;
use std::os::{
    fd::FromRawFd,
    unix::process::{parent_id, CommandExt},
};
use std::process::{exit, Command};
use std::time::Duration;

use anyhow::{bail, Context};
use clap::Parser;
use nix::unistd::{fork, ForkResult};
use tokio::time::timeout;
use zbus::{export::futures_util::stream::StreamExt, fdo::DBusProxy, names::BusName, Connection};

#[derive(Parser)]
#[clap(author)]
struct Cli {
    #[arg(short, long)]
    notify_fd: i32,

    #[arg(short, long)]
    bus_name: String,

    #[arg(short, long, default_value = "10")]
    timeout_secs: u64,

    cmd: Vec<OsString>,
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let cli = Cli::parse();

    let conn = Connection::session()
        .await
        .context("connect to user d-bus")?;
    let proxy = DBusProxy::new(&conn).await.context("creating dbus proxy")?;
    let mut owner_changed = proxy
        .receive_name_owner_changed_with_args(&[(0, &cli.bus_name)])
        .await?;

    // SAFETY: This is only unsafe for multithreaded programs. This program is single-thread.
    match unsafe { fork()? } {
        ForkResult::Parent { child: _ } => {
            // TODO: I could pipe this program's stdin to an fd passed to the child.
            //       The child can then assert whether the parent is still alive.
            let mut args = cli.cmd.iter();
            let cmd = args.next().context("must specify a command to run")?;
            // Only returns if 'exec' calls fails:
            anyhow::bail!(Command::new(cmd).args(args).exec());
        }
        ForkResult::Child => {
            let parent_pid = parent_id();

            // Double fork so process gets reaped even if service doesn't reap unknown children.
            // SAFETY: This is only unsafe for multithreaded programs. This program is single-thread.
            if unsafe { fork()? }.is_parent() {
                exit(0);
            };

            // SAFETY: no other functions should call `from_raw_fd`, so there
            // is only one owner for the file descriptor.
            let mut notify_fd = unsafe { File::from_raw_fd(cli.notify_fd) };

            timeout(Duration::from_secs(cli.timeout_secs), {
                async move {
                    while let Some(event) = owner_changed.next().await {
                        if let Some(new_owner) = event.args()?.new_owner.take() {
                            let new_owner = proxy
                                .get_connection_unix_process_id(BusName::Unique(new_owner))
                                .await
                                .context("getting new owner pid")?;

                            if new_owner != parent_pid {
                                bail!("another process took over the socket");
                            }

                            notify_fd
                                .write_all(b"READY=1\n")
                                .context("writing to readiness fd")?;
                            exit(0);
                        }
                    }
                    unreachable!("stream of owner changed events should never end");
                }
            })
            .await??;
            unreachable!("main loop must exit or time out");
        }
    }
}
